import sys
import subprocess
import shlex


def get_text():
    if sys.stdin.isatty():
        text = " ".join(sys.argv[2:]).strip()
    else:
        text = sys.stdin.read()
    return text


def get_command(override = ""):
    if override: return shlex.split(override)[0]
    return sys.argv[1]

def print_options():
    print("Options:")
    print("span: ".ljust(16)+"The records to select, example: span:7d is the last 7 days.")
    print("format: ".ljust(16)+"The format to see the records in. Only valid for show currently. [html, json, short]")
    print("@{tag}: ".ljust(16)+"This tag or any others must be included, example: @notes @school, records must have either.")
    print("+@{tag}: ".ljust(16)+"This tag and any others must be included. i.e. +@notes +@school, records must have both.")
    print("-@{tag}: ".ljust(16)+"This tag must not be included. i.e. -@notes @school, records for school, no notes.")
    print("{text}: ".ljust(16)+"This text is optional.")
    print("+{text}: ".ljust(16)+"This text must be included.")
    print("-{text}: ".ljust(16)+"This text must not be included")
    print("\nNote, text must be quoted if there are spaces.")


def parse_options(override = ""):
    if override: splits = shlex.split(override)[1:]
    else: splits = sys.argv[2:]
    ocontains = []
    ncontains = []
    acontains = []
    atags = []
    otags = []
    ntags = []
    span = None
    format = ""
    for split in splits:
        if split.startswith("span:"): span = parse_span(split[len("span:"):].split(",",maxsplit=2))
        elif split.startswith("format:"): format = split[len("format:"):].lower()
        elif split.startswith("+@"): atags.append(split[2:].lower())
        elif split.startswith("-@"): ntags.append(split[2:].lower())
        elif split.startswith("@"): otags.append(split[1:].lower())
        elif split.startswith("+"): acontains.append(split[1:].lower())
        elif split.startswith("-"): ncontains.append(split[1:].lower())
        else: ocontains.append(split.lower())

    return {"otags":otags, "atags":atags, "ntags":ntags,
            "acontains":acontains, "ocontains":ocontains, "ncontains":ncontains,
            "span":span, "format":format}


def parse_span(args):
    from datetime import datetime, timedelta
    import re
    now = datetime.now()
    operations = []
    
    for select in args:
        select = select.strip()
        try:
            select = datetime.fromisoformat(select)
            operations.append(select)
        except ValueError:
            units = [re.search("(-?\d+)"+u, select, re.IGNORECASE) for u in ["y","w","d","h","m","s"]]
            units = [int(u.group(1)) if u else 0 for u in units]
            days = units[0]*365+units[1]*7+units[2]
            seconds = units[3]*(60*60)+units[4]*60+units[5]

            if days or seconds:
                if operations:
                    if isinstance(operations[0], int):
                        operations.append(timedelta(days=days, seconds=seconds))
                    else:
                        operations.append(operations[0]+timedelta(days=days, seconds=seconds))
                else:
                    days = abs(days)*-1
                    seconds = abs(seconds)*-1
                    operations.append(now+timedelta(days=days, seconds=seconds))
            else:
                try:
                    num = int(select)
                    if operations:
                        operations.append(num)
                    else:
                        operations.append(abs(num)*-1)
                except ValueError:
                    print("bad record range format: "+select)
                    pass

    if len(operations) == 1: operations.append(now)
    return operations


def get_safe_filename():
    return "".join([c for c in "_".join(sys.argv[1:]) if c.isalpha() or c.isdigit() or c in ['_']]).rstrip()


def run(text):
    print("running "+text)
    subprocess.run(text, shell=True)


def popen(text):
    print("running "+text)
    return subprocess.Popen(text, shell=True)


import os
import platform
import time
import tempfile
import tdb.cli
import tdb.tags
import tdb.config
import tdb.records
import signal
import sys

_last_text = ""
_start_text = ""

def start(name, content="", ext=".txt", update_cb=None):
    global _last_text
    global _start_text
    _start_text = content
    name += "-"
    temp_file = tempfile.NamedTemporaryFile(mode="w+", prefix=name, suffix=ext, delete=False)
    temp_file.write(content)
    temp_file.close()
    proc = tdb.cli.popen(f"{tdb.config.get('editor')} {temp_file.name}")
    modtime = os.path.getmtime(temp_file.name)
    try:
        while proc.poll() == None:
            if modtime < os.path.getmtime(temp_file.name):
                time.sleep(0.1)
                modtime = os.path.getmtime(temp_file.name)
                text = open(temp_file.name).read()
                text = tdb.tags.parse_cmds("update", text)
                _last_text = text
                if update_cb:
                    text = update_cb(content, text)
                    content = text
                open(temp_file.name, "w").write(text)
                modtime = os.path.getmtime(temp_file.name)
            time.sleep(0.1)
    except KeyboardInterrupt as e:
        print("\nInterrupt detected!")
    
    text = ""
    if os.path.exists(temp_file.name):
        text = open(temp_file.name).read()
        os.remove(temp_file.name)
    return text

 
def signal_term_handler(signal, frame):
    if _last_text != _start_text:
        tdb.records.add_record(_last_text)
    sys.exit(1)
 
signal.signal(signal.SIGTERM, signal_term_handler)
if platform.system() == "Linux": signal.signal(signal.SIGQUIT, signal_term_handler)


import os
import atexit
import tdb.config

_skip_shutdown = True
_db_file = tdb.config.get("db_file")
_db_archive = tdb.config.get("db_archive")
_db_merge_func = None
_db_has_conflicts = False

if not _db_file:
    _db_file = tdb.config.get_tdb_dir()
    _db_file = os.path.join(_db_file, "db.txt")


if not os.path.exists(_db_file):
    from urllib.request import Request, urlopen
    import json
    first_line = "oh hai\n"
    try:
        # At some point in the future people will want this removed. :)
        request = Request('https://icanhazdadjoke.com/', headers={'Accept': 'application/json', 'User-Agent': "tdb"})
        response = urlopen(request).read().decode()
        response = json.loads(response)
        if "joke" in response: first_line = response["joke"]+"\n"
    except Exception as e: pass
    open(_db_file, "w").write(first_line)


if not _db_archive:
    _db_archive = tdb.config.get_tdb_dir()
    _db_archive = os.path.join(_db_archive, "db_archive.txt")


if not os.path.exists(_db_archive):
    open(_db_archive, "w").write("archived entries go here.")


_db_mtime = os.path.getmtime(_db_file)
_db_text = ""
_db_inserts = []

def get_mtime(): return _db_mtime

def _init():
    global _db_text
    global _skip_shutdown
    if not _db_text:
        _db_text = open(_db_file).read()
        _skip_shutdown = False


def get_filename(): return _db_file
def get_archive(): return _db_archive
def get_text():
    _init()
    return _db_text


def set_text(text):
    global _db_text
    _init()
    _db_text = text


def append(text):
    global _db_inserts
    global _db_text
    _init()
    if _db_text and _db_text[-1] != '\n':
        text = '\n'+text
    insert(text, len(_db_text), len(_db_text))

def append_immediate(text): open(get_filename(), "a").write(text)

def replace(old, new):
    global _db_inserts
    _init()
    # print("old:"+str([old]))
    id = _db_text.index(old)
    if id != -1:
        insert(new, id, id+len(old))

def archive(text, remove=True):
    open(get_archive(), "a").write(text)
    if remove: replace(text, "")


def insert(text, start, end):
    global _db_inserts
    _init()
    _db_inserts.append([text, (start, end)])


def perform_inserts():
    global _db_text
    global _db_inserts
    # print(_db_inserts)
    while _db_inserts:
        insert, span =_db_inserts.pop(0)
        delta = len(insert)-(span[1]-span[0])
        # print(delta)
        # print(len(insert))
        # print(_db_text[:span[0]])
        # print("-")
        # print(insert)
        # print("-")
        # print(_db_text[span[1]:])
        _db_text = _db_text[:span[0]] + insert + _db_text[span[1]:]
        _db_inserts = [[i[0], (i[1][0]+delta, i[1][1]+delta)] if i[1][0] >= span[0] else i for i in _db_inserts]
        # print(_db_inserts)
        # print("----------------------------------")

    pass


def serialise():
    global _db_text
    global _db_inserts
    global _db_mtime
    global _db_has_conflicts

    db_head = _db_text if _db_inserts else ""
    perform_inserts()

    if db_head and _db_mtime != os.path.getmtime(_db_file):
        output = _db_merge_func(db_head, _db_text, open(get_filename()).read())
        open(get_filename(), "w").write(output)
        _db_text = output
        if _db_has_conflicts:
            print("@tdb_conflict has been added to affected records.")
    elif db_head:
        open(get_filename(), "w").write(_db_text)
    _db_mtime = os.path.getmtime(_db_file)
    _db_has_conflicts = False


@atexit.register
def _shutdown():
    if _skip_shutdown: return
    serialise()


# Implementation of RAKE - Rapid Automtic Keyword Exraction algorithm
# as described in:
# Rose, S., D. Engel, N. Cramer, and W. Cowley (2010). 
# Automatic keyword extraction from indi-vidual documents. 
# In M. W. Berry and J. Kogan (Eds.), Text Mining: Applications and Theory.unknown: John Wiley and Sons, Ltd.
# @Graham Hughes, optimisations.

import re
from difflib import SequenceMatcher

def similarity_score(a, b):
    return SequenceMatcher(None, a, b).ratio()

_stop_words =["a", "a's", "able", "about", "above", "according", "accordingly", "across", "actually", "after", "afterwards", "again",
              "against", "ain't", "all", "allow", "allows", "almost", "alone", "along", "already", "also", "although", "always", "am",
              "among", "amongst", "an", "and", "another", "any", "anybody", "anyhow", "anyone", "anything", "anyway", "anyways", "anywhere",
              "apart", "appear", "appreciate", "appropriate", "are", "aren't", "around", "as", "aside", "ask", "asking", "associated",
              "at", "available", "away", "awfully", "b", "be", "became", "because", "become", "becomes", "becoming", "been", "before",
              "beforehand", "behind", "being", "believe", "below", "beside", "besides", "best", "better", "between", "beyond", "both",
              "brief", "but", "by", "c", "c'mon", "c's", "came", "can", "can't", "cannot", "cant", "cause", "causes", "certain", "certainly",
              "changes", "clearly", "co", "com", "come", "comes", "concerning", "consequently", "consider", "considering", "contain",
              "containing", "contains", "corresponding", "could", "couldn't", "course", "currently", "d", "definitely", "described",
              "despite", "did", "didn't", "different", "do", "does", "doesn't", "doing", "don't", "done", "down", "downwards", "during",
              "e", "each", "edu", "eg", "eight", "either", "else", "elsewhere", "enough", "entirely", "especially", "et", "etc", "even",
              "ever", "every", "everybody", "everyone", "everything", "everywhere", "ex", "exactly", "example", "except", "f", "far",
              "few", "fifth", "first", "five", "followed", "following", "follows", "for", "former", "formerly", "forth", "four", "from",
              "further", "furthermore", "g", "get", "gets", "getting", "given", "gives", "go", "goes", "going", "gone", "got", "gotten",
              "greetings", "h", "had", "hadn't", "happens", "hardly", "has", "hasn't", "have", "haven't", "having", "he", "he's", "hello",
              "help", "hence", "her", "here", "here's", "hereafter", "hereby", "herein", "hereupon", "hers", "herself", "hi", "him", "himself",
              "his", "hither", "hopefully", "how", "howbeit", "however", "i", "i'd", "i'll", "i'm", "i've", "ie", "if", "ignored", "immediate",
              "in", "inasmuch", "inc", "indeed", "indicate", "indicated", "indicates", "inner", "insofar", "instead", "into", "inward",
              "is", "isn't", "it", "it'd", "it'll", "it's", "its", "itself", "j", "just", "k", "keep", "keeps", "kept", "know", "knows",
              "known", "l", "last", "lately", "later", "latter", "latterly", "least", "less", "lest", "let", "let's", "like", "liked", "likely",
              "little", "look", "looking", "looks", "ltd", "m", "mainly", "many", "may", "maybe", "me", "mean", "meanwhile", "merely", "might",
              "more", "moreover", "most", "mostly", "much", "must", "my", "myself", "n", "name", "namely", "nd", "near", "nearly", "necessary",
              "need", "needs", "neither", "never", "nevertheless", "new", "next", "nine", "no", "nobody", "non", "none", "noone", "nor", "normally",
              "not", "nothing", "novel", "now", "nowhere", "o", "obviously", "of", "off", "often", "oh", "ok", "okay", "old", "on", "once", "one",
              "ones", "only", "onto", "or", "other", "others", "otherwise", "ought", "our", "ours", "ourselves", "out", "outside", "over",
              "overall", "own", "p", "particular", "particularly", "per", "perhaps", "placed", "please", "plus", "possible", "presumably",
              "probably", "provides", "q", "que", "quite", "qv", "r", "rather", "rd", "re", "really", "reasonably", "regarding", "regardless",
              "regards", "relatively", "respectively", "right", "s", "said", "same", "saw", "say", "saying", "says", "second", "secondly", "see",
              "seeing", "seem", "seemed", "seeming", "seems", "seen", "self", "selves", "sensible", "sent", "serious", "seriously", "seven", "several",
              "shall", "she", "should", "shouldn't", "since", "six", "so", "some", "somebody", "somehow", "someone", "something", "sometime",
              "sometimes", "somewhat", "somewhere", "soon", "sorry", "specified", "specify", "specifying", "still", "sub", "such", "sup",
              "sure", "t", "t's", "take", "taken", "tell", "tends", "th", "than", "thank", "thanks", "thanx", "that", "that's", "thats",
              "the", "their", "theirs", "them", "themselves", "then", "thence", "there", "there's", "thereafter", "thereby", "therefore",
              "therein", "theres", "thereupon", "these", "they", "they'd", "they'll", "they're", "they've", "think", "third", "this",
              "thorough", "thoroughly", "those", "though", "three", "through", "throughout", "thru", "thus", "to", "together", "too", "took",
              "toward", "towards", "tried", "tries", "truly", "try", "trying", "twice", "two", "u", "un", "under", "unfortunately", "unless",
              "unlikely", "until", "unto", "up", "upon", "us", "use", "used", "useful", "uses", "using", "usually", "uucp", "v", "value", "various",
              "very", "via", "viz", "vs", "w", "want", "wants", "was", "wasn't", "way", "we", "we'd", "we'll", "we're", "we've", "welcome", "well",
              "went", "were", "weren't", "what", "what's", "whatever", "when", "whence", "whenever", "where", "where's", "whereafter", "whereas",
              "whereby", "wherein", "whereupon", "wherever", "whether", "which", "while", "whither", "who", "who's", "whoever", "whole", "whom",
              "whose", "why", "will", "willing", "wish", "with", "within", "without", "won't", "wonder", "would", "would", "wouldn't", "x", "y",
              "yes", "yet", "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves", "z", "zero"]

def is_number(s):
    try:
        float(s) if '.' in s else int(s)
        return True
    except ValueError:
        return False


def load_stop_words(stop_word_file):
    """
    Utility function to load stop words from a file and return as a list of words
    @param stop_word_file Path and file name of a file containing stop words.
    @return list A list of stop words.
    """
    stop_words = []
    for line in open(stop_word_file):
        if line.strip()[0:1] != "#":
            for word in line.split():  # in case more than one per line
                stop_words.append(word)
    return stop_words


def separate_words(text, min_word_return_size):
    """
    Utility function to return a list of all words that are have a length greater than a specified number of characters.
    @param text The text that must be split in to words.
    @param min_word_return_size The minimum no of characters a word must have to be included.
    """
    splitter = re.compile('[^a-zA-Z0-9_\\+\\-/]')
    words = []
    for single_word in splitter.split(text):
        current_word = single_word.strip().lower()
        #leave numbers in phrase, but don't count as words, since they tend to invalidate scores of their phrases
        if len(current_word) > min_word_return_size and current_word != '' and not is_number(current_word):
            words.append(current_word)
    return words


def split_sentences(text):
    """
    Utility function to return a list of sentences.
    @param text The text that must be split in to sentences.
    """
    sentence_delimiters = re.compile(u'[.!?,;:\t\\\\"\\(\\)\\\'\u2019\u2013]|\\s\\-\\s')
    sentences = sentence_delimiters.split(text)
    return sentences


def build_stop_word_regex(stop_word_file_path):
    stop_word_list = load_stop_words(stop_word_file_path) if stop_word_file_path else _stop_words
    stop_word_regex_list = []
    for word in stop_word_list:
        word_regex = r'\b' + word + r'(?![\w-])'  # added look ahead for hyphen
        stop_word_regex_list.append(word_regex)
    stop_word_pattern = re.compile('|'.join(stop_word_regex_list), re.IGNORECASE)
    return stop_word_pattern


def generate_candidate_keywords(sentence_list, stopword_pattern):
    phrase_list = []
    for s in sentence_list:
        tmp = re.sub(stopword_pattern, '|', s.strip())
        phrases = tmp.split("|")
        for phrase in phrases:
            phrase = phrase.strip().lower()
            if phrase != "":
                phrase_list.append(phrase)
    return phrase_list


def calculate_word_scores(phraseList):
    word_frequency = {}
    word_degree = {}
    for phrase in phraseList:
        word_list = separate_words(phrase, 0)
        word_list_length = len(word_list)
        word_list_degree = word_list_length - 1
        #if word_list_degree > 3: word_list_degree = 3 #exp.
        for word in word_list:
            word_frequency.setdefault(word, 0)
            word_frequency[word] += 1
            word_degree.setdefault(word, 0)
            word_degree[word] += word_list_degree  #orig.
            #word_degree[word] += 1/(word_list_length*1.0) #exp.
    for item in word_frequency:
        word_degree[item] = word_degree[item] + word_frequency[item]

    # Calculate Word scores = deg(w)/frew(w)
    word_score = {}
    for item in word_frequency:
        word_score.setdefault(item, 0)
        word_score[item] = word_degree[item] / (word_frequency[item] * 1.0)  #orig.
    #word_score[item] = word_frequency[item]/(word_degree[item] * 1.0) #exp.
    return word_score


def generate_candidate_keyword_scores(phrase_list, word_score):
    keyword_candidates = {}
    for phrase in phrase_list:
        keyword_candidates.setdefault(phrase, 0)
        word_list = separate_words(phrase, 0)
        candidate_score = 0
        for word in word_list:
            candidate_score += word_score[word]
        keyword_candidates[phrase] = candidate_score
        phrase_words = phrase.split()
        split_score = (candidate_score/len(phrase_words))
        if split_score > 1:
            # input(phrase_words)
            for word in phrase_words:
                if word in keyword_candidates:
                    keyword_candidates[word] += split_score
                else:
                    keyword_candidates[word] = split_score
    return keyword_candidates


class Rake(object):
    def __init__(self, stop_words_path=""):
        self.stop_words_path = stop_words_path
        self.__stop_words_pattern = build_stop_word_regex(stop_words_path)

    def run(self, text):
        sentence_list = split_sentences(text)

        phrase_list = generate_candidate_keywords(sentence_list, self.__stop_words_pattern)

        word_scores = calculate_word_scores(phrase_list)

        keyword_candidates = generate_candidate_keyword_scores(phrase_list, word_scores)

        sorted_keywords = sorted(keyword_candidates.items(), key=lambda x: x[1], reverse=True)
        return sorted_keywords


import re

re_tag = re.compile(r'\s@(\w+)')
_cmd_tags = {}

def _safe_re_search(string, position, pattern) -> int:
    match = pattern.search(string, position)
    return match.span()[0] if match else -1


def find_tags(text: str):
    #edge case, text can't start with @
    re_end = re.compile(r'([\r\n])')
    tags = []
    tag_spans = []
    for match in re_tag.finditer(text):
        tag = match.group(1)
        end = match.span()[1]
        match_start = match.span()[0]
        skip = False

        for span in tag_spans:
            if span[0] < match_start and span[1] > match_start:
                skip = True
                break

        if not skip and len(text) > end+1 and text[end] == ':':
            start = end+1
            end = _safe_re_search(text, start, re_end)
            if end == -1: end = len(text)
            tag_spans.append((start, end))
            tags.append((tag, text[start:end].strip()))
            
        elif not skip:
            tags.append((tag, ""))

    return tags

def replace_tag(text: str, tag, repl):
    #edge case, text can't start with @
    if tag[1]:
        pattern = "\s?@"+tag[0]+":\s*"+re.escape(tag[1])
    else:
        pattern = "\s?@"+tag[0]+"(?!\S)"
    re_sub_tag = re.compile(pattern)
    return re_sub_tag.sub(repl, text)


def contains_tag(text, tag):
    text = text.lower()
    tag = tag.lower()
    for match in re_tag.finditer(text):
        if match.group(1) == tag:
            return True
    return False


def register_cmd(tag, func):
    print(f"Registering '@{tag}' cmds.")
    _cmd_tags[tag] = func


def parse_cmds(context, text):
    tags = find_tags(text)
    for tag in tags:
        if tag[0] in _cmd_tags:
            text = _cmd_tags[tag[0]](context, text, tag[1])
    return text




import tdb.config
import sys
try: import markdown
except: pass
try: import md_mermaid
except: pass
import os

_css_file = "/".join((tdb.config._tdb_dir, "style.css"))

if not os.path.exists(_css_file): _css_file = "/".join((os.path.dirname(__file__), "style.css"))

_css = open(_css_file, "r").read()

body = """<html>
      <header>
         <script src="mermaid.min.js"></script>
        <style>
/*inserted from {css_file}*/
{css}
        </style>
    </header>
    <body>
    <div class="entry_spacer"></div>
    <div id="container" class="container">
        <div class="entry_spacer"></div>
{entries}
    </div>
    </body>
</html>
"""
entry = """
    <div class="entry">
        <div class="date">
{date}
        </div>
        <div class="content">
{text}
        </div>
    </div>
    <div class="entry_spacer"></div>
"""

def print_html(entries, file=None):
    entries_str = ""
    for in_entry in entries:
        if "markdown" in sys.modules:
            extensions = ["extra", "codehilite"]
            if "md_mermaid" in sys.modules: extensions.append("md_mermaid")
            try: in_entry["text"] = markdown.markdown(in_entry["text"], extensions=extensions)
            except: in_entry["text"] = markdown.markdown(in_entry["text"], extensions=["extra", "codehilite"])
        else:
            in_entry["text"] = "<pre>"+in_entry["text"]+"</pre>"
        entries_str += entry.format_map(in_entry)
    print(body.format_map({"css":_css, "css_file":_css_file, "entries":entries_str}), file=file)



import re
from datetime import datetime
import time
import json
import tdb.db
import tdb.tags
import tdb.rake
import tdb.html
import tdb.cli
# This is the format: "2023-04-05 09:59:33"
re_iso_record = re.compile(r'^\[tdb:(\d{4}\-\d{2}-\d{2} \d{2}:\d{2}:\d{2}(\.\d{6})?)\] ?', re.MULTILINE | re.IGNORECASE)
re_hex_record = re.compile(r'^\[tdb:(0x[\da-f]+)\] ?', re.MULTILINE | re.IGNORECASE)
_record_cache = []
_record_cmds = []
_needs_sort = False
_force_hex = False
_record_mtime = 0

def convert_db_nano_to_micro():
    return # don't call this code atm.
    file_name = tdb.db.get_filename()
    text = open(file_name, "r").read()
    for match in re_hex_record.finditer(text):
        nano = int(match.group(1), 16)
        if int(nano/1E9) > 1E9:
            before = f"[tdb:{(hex(nano))}]"
            nano = int(nano/1E3)
            after = f"[tdb:{(hex(nano))}]"
            print(f"before: {before}\nafter: {after}")
            text = text.replace(before, after, 1)
    open(file_name, "w").write(text)
    

def convert_headers(text):
    out = text
    while m := re_iso_record.search(out):
        d = datetime.fromisoformat(m.group(1))
        d = hex(int(d.timestamp()*1E6))
        out = out[:m.span(1)[0]]+d+out[m.span(1)[1]:]
    return out


class Record(object):
    text = ""
    time = 0
    delta = 0
    date = None
    tags = []
    span = (0,0)
    pos = 0
    id = -1
    def __init__(self, text, time, delta, date, tags, span, pos, id):
        self.text = text
        if isinstance(date, str):
            self.date = datetime.fromisoformat(date)
        else:
            self.date = date
        self.tags = tags
        self.span = span
        self.time = time
        self.delta = delta
        self.id = id
        self.pos = pos

    def __str__(self):
        if _force_hex:
            return self.entry()
        else:
            return f"[tdb:{self.iso_str()}] {self.text}"

    def iso_str(self): return self.date.isoformat(' ')

    def entry(self):
        return f"[tdb:{hex(self.time)}] {self.text}"
    
    def asdict(self):
        return {'text': self.text, 'time': self.time, 'date': self.date.isoformat(" "), 'tags': self.tags, 'span':self.span, 'id':self.id }


def register_cmd(func):
    global _record_cmds
    _record_cmds.append(func)
    pass


def make_record(date, text):
    return f"\n[tdb:{date}] {text}"


def add_record(text):
    dt = tdb.db.get_mtime()
    ns = time.time_ns()
    # convert to microseconds for datetime compliance
    ns = int(ns/1E3)

    if (ns/1E6) - dt > 1.0: ns = int(int(ns/1E6)*1E6)
    # if (ns/1E9) - dt > 1.0: ns = int(int(ns/1E9)*1E9)
    record = make_record(hex(ns), text)
    tdb.db.append_immediate(record)
    tdb.db.archive(record, False)

    print("Record added successfully!")
    for r in _record_cmds: r(text)


def deduplicate_records(in_records):
    dedupe = []
    for r1 in in_records:
        new = True
        for r2 in dedupe:
            if r1.date == r2.date: new = False; break
        if new: dedupe.append(r1)
    return dedupe


def modify_db_records(previous, current):
    records = split_records(previous)
    new_records = split_records(current)
    dedupe = deduplicate_records(new_records)
    if len(new_records) != len(dedupe): print("Warning: duplicate dates found. Ignoring those entries.")
    new_records = dedupe
    found = []
    mods = adds = dels = 0
    for r1 in new_records:
        modified = None
        new = True
        for r2 in records:
            if str(r1) != str(r2) and r1.iso_str() == r2.iso_str():
                modified = r2
                found.append(r2)
                new = False
                break
            elif r1.iso_str() == r2.iso_str():
                found.append(r2)
                new = False
                break
        
        if modified:
            tdb.db.replace(r2.entry(), r1.entry()); mods += 1
        elif new:
            # print(f"new: {r1}")
            # for r2 in records:
            #     print((r1.iso_str()," != ",r2.iso_str()))
            tdb.db.append(r1.entry()); adds+=1
            pass
    for r1 in records:
        if not r1 in found:
            # print(f"del: {r1}")
            tdb.db.archive(r1.entry()); dels+=1
            pass
    
    if adds or mods or dels: print("".ljust(32, "="))
    if adds: print(f"Inserted {adds} record{'s' if adds > 1 else ''}.")
    if mods: print(f"Modified {mods} record{'s' if mods > 1 else ''}.")
    if dels: print(f"Archived {dels} record{'s' if dels > 1 else ''}.")
    
    for r in _record_cmds: r(current)


def merge_records(text_head, text_a, text_b):
    head = split_records(text_head)
    a_db = split_records(text_a)
    b_db = split_records(text_b)
    def handle_conflict(in_record):
        if not tdb.db._db_has_conflicts:
            print("Warning: conflict in ")
            tdb.db._db_has_conflicts = True
        print("\t"+str(in_record).splitlines()[0])
        in_record.text += "\n@tdb_conflict\n\n"
        
    out = []
    while head or a_db or b_db:
        h = head.pop(0) if head else None
        a = a_db.pop(0) if a_db else None
        b = b_db.pop(0) if b_db else None
        # if h: print("h:"+h.entry().splitlines()[0])
        # if a: print("a:"+a.entry().splitlines()[0])
        # if b: print("b:"+b.entry().splitlines()[0])
        # do something so these eventually all reference the same thing?
        if h and a and b and (h.date < a.date or h.date < b.date): 
            while head and h.date < a.date or h.date < b.date:
                if h.date == a.date and h.text != a.text: # change
                    handle_conflict(a)
                    out.append(a)
                elif h.date == b.date and h.text != b.text: # change
                    handle_conflict(b)
                    out.append(b)

                h = head.pop(0)

            if a.date < b.date:
                while a_db and a.date != b.date: a = a_db.pop(0)
            else:
                while b_db and a.date != b.date: b = b_db.pop(0)

        if h and a and b and (h.date == a.date and h.date == b.date):
            if h.text != a.text and h.text == b.text: h.text = a.text
            elif h.text == a.text and h.text != b.text: h.text = b.text
            elif h.text != a.text and h.text != b.text:
                import difflib
                diff = difflib.Differ().compare(a.text.splitlines(keepends=True), b.text.splitlines(keepends=True))
                h.text = "".join([l[2:] for l in list(diff)])
                handle_conflict(h)
            out.append(h)
            continue
        # don't lose this record
        if a: a_db.append(a)
        if b: b_db.append(b)
        if h: head.append(h)
        break
    
    out = out + head + a_db + b_db
    out = deduplicate_records(out)
    sorted(out, key=lambda x: x.date)
    out = "".join([r.entry() for r in out])
    # print(len(out.splitlines()))
    return out


tdb.db._db_merge_func = merge_records

def print_records(options=None):
    results = []
    results = split_db_records(options)
    if options and options["format"] == "json":
        res = [r.asdict() for r in results]
        print(json.dumps(res, indent=2))
    elif options and options["format"] == "html":
        tdb.html.print_html(reversed([r.asdict() for r in results]))
    elif options and options["format"] == "short":
        out = "".join([str(r).splitlines()[0]+"\n" for r in results])
        print(out.strip())
    else:
        out = "".join([str(r) for r in results])
        print(out.strip())


def filter_records(records, options):
    max_id = len(records)
    id_offset = -1
    out = []
    end_date = None

    span = options["span"] if options else []
    otags = options["otags"] if options else []
    ntags = options["ntags"] if options else []
    atags = options["atags"] if options else []
    acontains = options["acontains"] if options else []
    ocontains = options["ocontains"] if options else []
    ncontains = options["ncontains"] if options else []

    # TODO: id's should be linear by this point
    # optimizations like:
    # > if all(map(lambda x: isinstance(x, int), span)): records[span[0]:span[1]]
    # should be very possible.
    # for date ranged, just break when you leave the range.

    for record in records:
        date = record.date
        low_text = record.text.lower()
        skip = False
        if not skip and span:
            skip = True

            if all(map(lambda x: isinstance(x, int), span)):
                skip = not (max_id+span[0] < record.id <= max_id+span[0]+span[1])

            elif isinstance(span[0], int):
                if span[0] < 0 and (max_id+span[0]) < record.id:
                    if not end_date:
                        if isinstance(span[1], datetime): end_date = span[1]
                        else: end_date = record.date + span[1]
                    skip = not (date <= end_date)
                else: assert("we're not doing dates in the future")

            elif isinstance(span[1], int):

                if span[1] >= 0 and date >= span[0]:
                    if id_offset == -1: id_offset = record.id
                    skip = not ((record.id - id_offset) < span[1])
                
                elif span[1] < 0 and date <= span[0]:
                    if id_offset == -1: id_offset = record.id
                    skip = not (abs(record.id-id_offset) < abs(span[1]))
                
                else:
                    skip = span[1] >= 0 and date <= span[0]

            elif span[0] <= date <= span[1]: skip = False

        if not skip and ocontains: skip = not any([c in low_text for c in ocontains])
        if not skip and acontains: skip = not all([c in low_text for c in acontains])
        if not skip and ncontains: skip = any([c in low_text for c in ncontains])
        if not skip and otags: skip = not any([tdb.tags.contains_tag(low_text, t) for t in otags])
        if not skip and atags: skip = not all([tdb.tags.contains_tag(low_text, t) for t in atags])
        if not skip and ntags: skip = any([tdb.tags.contains_tag(low_text, t) for t in ntags])
        if not skip: out.append(record)
    
    return out


def sort_db_records():
    global _record_cache
    global _needs_sort
    print("warning: db unordered, attempting sort.")
    tdb.db.set_text("".join([r.entry() for r in _record_cache]))
    _needs_sort = False
    _record_cache = split_records(tdb.db.get_text())
    assert(_needs_sort == False)


def split_db_records(options=None):
    global _record_cache
    global _needs_sort
    global _record_mtime
    if tdb.db._db_mtime != _record_mtime:
        _record_cache = split_records(tdb.db.get_text())
        _record_mtime = tdb.db._db_mtime
    if _needs_sort: sort_db_records()
    if options: return filter_records(_record_cache, options)
    else: return _record_cache


def split_records(text: str, options=None):
    global _needs_sort
    text = convert_headers(text)
    last = None
    current = None
    id = 0
    id_offset = -1

    filtered = []

    otags = options["otags"] if options else []
    ntags = options["ntags"] if options else []
    atags = options["atags"] if options else []
    span = options["span"] if options else []
    acontains = options["acontains"] if options else []
    ocontains = options["ocontains"] if options else []
    ncontains = options["ncontains"] if options else []

    def append_record():
        nonlocal filtered
        nonlocal last
        nonlocal current
        nonlocal id
        nonlocal id_offset

        if last:
            id += 1
            x,y = last["span"][1], current["span"][0]
            section = text[x:y]
            date = last["date"]
            last["text"] = section
            last["span"] = (x ,y)
            last["id"] = id


            skip = False
            if not skip and span:
                skip = True

                if all(map(lambda x: isinstance(x, int), span)): skip = False
                elif isinstance(span[0], int): skip = False
                elif isinstance(span[1], int):

                    if span[1] >= 0 and date >= span[0]:
                        if id_offset == -1: id_offset = id
                        skip = not (id - id_offset) < span[1]
                    else:
                        skip = span[1] >= 0 and date <= span[0]

                elif span[0] <= date <= span[1]: skip = False

            sec_low = section.lower()

            if not skip and ocontains: skip = not any([c in sec_low for c in ocontains])
            if not skip and acontains: skip = not all([c in sec_low for c in acontains])
            if not skip and ncontains: skip = any([c in sec_low for c in ncontains])
            if not skip and otags: skip = not any([tdb.tags.contains_tag(sec_low, t) for t in otags])
            if not skip and atags: skip = not all([tdb.tags.contains_tag(sec_low, t) for t in atags])
            if not skip and ntags: skip = any([tdb.tags.contains_tag(sec_low, t) for t in ntags])
            if not skip: filtered.append(last)

    for match in re_hex_record.finditer(text):
        nano = int(match.group(1), 16)
        
        if int(nano/1E9) > 1E9:
            nano = int(nano/1E3)
        delta = 0
        if last: delta = nano - last["time"]
        current = {
            "date": datetime.fromtimestamp(nano/1E6),
            "time": nano,
            "delta": delta,
            "text": "",
            "id": 0,
            "tags": [],
            "span": match.span(),
            "pos": match.span()[0]
        }
        if last and last["time"] > current["time"]: _needs_sort = True

        append_record()
        last = current
    # hack to fix last record text
    current = {"span": [len(text)]}
    append_record()

    id_offset = -1
    end_date = None
    if span:
        # for record in records:
        def post_filter(record):
            nonlocal span
            nonlocal id_offset
            nonlocal id
            nonlocal end_date
            if all(map(lambda x: isinstance(x, int), span)):
                return id+span[0] < record["id"] <= id+span[0]+span[1]

            elif isinstance(span[1], int):
                # tested all the r[1] >= 0 above the given date.
                if span[1] >= 0: return True
                # TODO: fix this logic, 1d,-1 is valid and this doesn't get the right record
                elif span[1] < 0 and record["date"] <= span[0]:
                    if id_offset == -1: id_offset = record["id"]
                    return abs(record["id"]-id_offset) < abs(span[1])

            elif isinstance(span[0], int):
                if span[0] < 0 and (id+span[0]) < record["id"]:
                    if not end_date:
                        if isinstance(span[1], datetime): end_date = span[1]
                        else: end_date = record["date"] + span[1]

                    return record["date"] <= end_date
                
            else: return True
        
        filtered = filter(post_filter, filtered)
    if _needs_sort: filtered = sorted(filtered, key=lambda x: x["time"])
    filtered = [Record(**r) for r in filtered]
    return filtered


def find_similar(text):
    kw_ext = tdb.rake.Rake()
    text_kw = kw_ext.run(text)

    records = split_db_records()
    results = []
    for record in records:
        record_kw = kw_ext.run(record["text"])
        for k1, v1 in record_kw:
            for k2, v2 in text_kw:
                if tdb.rake.similarity_score(k1, k2) > 0.8:
                    record["score"] += (v1+v2)*0.5

        if record["score"] > 0:
            results.append(record)

    print(json.dumps(results, indent=2))


import shutil
import os

_editors = ["notepad", "gedit -w", "code -w", "emacs -a \"\" -c", "subl -w"]
for editor in _editors:
    editor = editor.split()[0]
    if shutil.which(editor):
        _editor = editor
        break

_tdb_dir = os.path.expanduser("~/.tdb")
_tdb_dir = _tdb_dir.replace("\\", "/")
_db_file = "/".join((_tdb_dir, "db.txt"))
_db_archive = "/".join((_tdb_dir, "db_archive.txt"))
_conf_file = "/".join((_tdb_dir, "config.toml"))
_addon_file = "addon.py"
_config = None


_conf_text = f"""\
# path to database
db_file = "{_db_file}"
db_archive = "{_db_archive}"
edit_ext = ".md"
# options: {_editors}
editor = "{_editor}" # command for editor
addons = ["{_addon_file}"]
"""

os.makedirs(_tdb_dir, exist_ok=True)

if not os.path.exists(_conf_file): open(_conf_file, "w").write(_conf_text)


def get_tdb_dir(): return _tdb_dir
def get_filename(): return _conf_file
def get(key):
    global _config
    if not _config:
        import tomllib
        _config = tomllib.load(open(_conf_file, "rb"))
    return _config.get(key)


import sys
import os
import tdb.tags
import tdb.records
import tdb.session
import tdb.config
import tdb.cli
import tdb.db
import importlib.util

_dirname = os.path.dirname(__file__)
_dirname = _dirname.replace("\\", "/")

def import_addon(file_path):
    basename = os.path.basename(file_path)
    if file_path == basename:
        file_path = "/".join((_dirname, file_path))
    if os.path.exists(file_path):
        basename, _ = os.path.splitext(basename)
        spec = importlib.util.spec_from_file_location("tdb.addon."+basename, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
    
    return None


def import_addons():

    if not tdb.config.get("addons"):
        print(f"No addons found in '{tdb.config.get_filename()}'")
        print("@tdb: commands will not work.")
        return

    for e in tdb.config.get("addons"):
        module = import_addon(e)
        if module:
            mod_vars = vars(module)
            if "get_addon_name" in mod_vars:
                if "addon_tag" in mod_vars:
                    tdb.tags.register_cmd(module.get_addon_name(), module.addon_tag)
                if "addon_record" in mod_vars:
                    tdb.records.register_cmd(module.addon_record)
            else:
                print("failed to add '"+str(e)+"'. missing expected interface 'get_addon_name'.")


def main():
    if len(sys.argv) < 2:
        print("# tdb\n\nA text based database with tagging.\n\n```")
        print("Usage: py -m tdb [add | edit | template | show | config | archive] [text | options]")
        print("".ljust(64,"-"))
        print("Commands:")
        print("add:".ljust(16)+"Make a record when text is supplied. Otherwise, open an editor to write one.")
        print("edit:".ljust(16)+"Open an editor with some view of the database, see options.")
        print("template:".ljust(16)+"Open an editor to write a record with the passed template file as a basis.")
        print("config:".ljust(16)+"Open tdbs config file.")
        print("archive:".ljust(16)+"Open tdbs archive. Check here if you lose a record.")
        print("".ljust(64,"-"))
        tdb.cli.print_options()
        print("```")
        sys.exit(1)
    command = tdb.cli.get_command()
    options = tdb.cli.parse_options()
    edit_ext = tdb.config.get('edit_ext')
    edit_ext = edit_ext if edit_ext else ".txt"
    if command == "add":
        import_addons()
        text = tdb.cli.get_text()
        if not text:
            text = tdb.session.start("tdb_add", ext=edit_ext)
        if text:
            tdb.records.add_record(text)
        else:
            print("No text provided. Record not added.")

    elif command == "show":
        tdb.records.print_records(options)

    elif command == "config":
        tdb.cli.run(f"{tdb.config.get('editor')} {tdb.config.get_filename()}")

    elif command == "archive":
            tdb.cli.run(f"{tdb.config.get('editor')} {tdb.db.get_archive()}")
            
    elif command == "edit":
        import_addons()
        if any(options.values()):
            records = tdb.records.split_db_records(options)
            content = "".join([str(r) for r in records])
            update_called = False
            def update_db(previous, text):
                nonlocal content
                nonlocal update_called
                update_called = True
                tdb.records.modify_db_records(previous, text)
                tdb.db.serialise()
                dates = [r.date for r in tdb.records.split_records(text)]
                text = "".join([str(r) for r in tdb.records.split_db_records() if r.date in dates])
                content = text
                return text

            text = tdb.session.start("tdb_"+tdb.cli.get_safe_filename(), content, ext=edit_ext, update_cb=update_db)
            if not update_called:
                print("no changes made")
                return
            elif content != text:
                update_db(content, text) # this should never happen.
        else:
            tdb.cli.run(f"{tdb.config.get('editor')} {tdb.db.get_filename()}")

    elif command == "template":
        import_addons()
        template = tdb.cli.get_text()
        if os.path.exists(template):
            basename, ext = os.path.splitext(template)
            basename = os.path.basename(basename)
            if not ext: ext = edit_ext
            content = open(template).read()
            text = tdb.session.start("tdb_"+basename, content, ext)
            if content == text:
                print("no changes made")
                return
            if text:
                tdb.records.add_record(text)
            else:
                print("No text provided. Record not added.")
        else:
            print(f"'{template}' is not a valid file")

    else:
        print("Invalid command. Try again.")
        sys.exit(1)
        
_profile = False
if __name__ == "__main__":
    if _profile:
        import cProfile
        cProfile.run("main()")
    else:
        main()


import shlex
import json
import tdb.tags
import tdb.records
import tdb.html

def get_addon_name(): return "tdb"

def addon_tag(context, text, args):
    print(context+" : "+str((get_addon_name(), args)))
    try:
        # TODO this used to be args.lower(), make sure removing that didn't break tags.
        split_args = shlex.split(args)
    except ValueError as e:
        split_args = []
    
    if split_args:
        if split_args[0] in ["add", "remove"]: text = tdb.tags.replace_tag(text, (get_addon_name(), args), "") 
        if split_args[0] == "remove": text = remove_tag_cmd(text, split_args[1:])
        elif split_args[0] == "add": text = add_tag_cmd(text, split_args[1:])
        elif split_args[0] == "export": text = export_cmd(text, split_args[1:])
    else:
        text = tdb.tags.replace_tag(text, (get_addon_name(), args), "")
    return text

def addon_record(text):
    pass

def add_tag_cmd(text, args):
    records = tdb.records.split_records(text)
    for r in records:
        for arg in args:
            if arg.startswith("@"):
                if not tdb.tags.contains_tag(r.text, arg[1:]):
                    lines = r.text.splitlines()
                    best = -1
                    for index in range(len(lines)-1, 0, -1):
                        if ":" not in lines[index] and lines[index].startswith("@"):
                            best = index
                            break

                    if best != -1:
                        if lines[best] and lines[best][-1] != " ": lines[best] += " "
                        lines[best] += arg
                        r.text = "\n".join(lines)+"\n"
                    else:
                        if len(lines[-1]) > 0: r.text += "\n"
                        r.text += "\n" + arg + "\n\n"

    return "".join([str(r) for r in records])


def remove_tag_cmd(text, args):
    tags = tdb.tags.find_tags(text)
    tags = filter(lambda x: "@"+x[0] in args, tags)
    for tag in tags:
        text = tdb.tags.replace_tag(text, tag, "")
    records = tdb.records.split_records(text)
    for r in records:
        r.text = r.text.rstrip() + "\n"

    return "".join([str(r) for r in records])


def export_cmd(text, args):
    path : str = args[0]
    out = "\n".join([l for l in text.splitlines() if not l.startswith("@tdb")])
    with open(path, "w+") as file:
        if path.endswith(".html"):
            #TODO this will allow multiple records with the same path to export together.
            #     however, this code runs for each export line, which is not performant.
            #     need a way to dedupe tag commands in some cases.
            records = [r for r in tdb.records.split_records(text) if path in r.text]
            for r in records: r.text = "\n".join([l for l in r.text.splitlines() if not l.startswith("@tdb")])
            tdb.html.print_html(reversed([r.asdict() for r in records]), file)
        elif path.endswith(".json"):
            records = [r for r in tdb.records.split_records(text) if path in r.text]
            for r in records: r.text = "\n".join([l for l in r.text.splitlines() if not l.startswith("@tdb")])
            res = [r.asdict() for r in records]
            print(json.dumps(res, indent=2), file=file)
        elif path.endswith(".short"):
            out = "".join([str(r).splitlines()[0]+"\n" for r in records])
            print(out.strip(), file=file)
        else:
            file.write(out)
    return text